"""
Adt for the Map Sort.
"""
# Aurelien Coet, 2018.

from alpynet.adt import Sort, GenericSort
from alpynet.adts.boolean import boolean


# Sort Definition.
kv_map = Sort('map')
generic = GenericSort()


# ---------- Operations on maps ---------- #
# Generators.
kv_map.operation('empty', ())
kv_map.operation('add', (kv_map, generic, generic))

# Observers.
kv_map.operation('get', (kv_map, generic))
kv_map.operation('isempty', (kv_map,), boolean)

# Modifiers.
kv_map.operation('delete', (kv_map, generic))

# Operations for the generic sort.
generic.operation('equal', (generic, generic))


# ---------- Variables ---------- #
kv_map.variable('m')
kv_map.variable('n')

# Generic variables for map keys.
generic.variable('k')
generic.variable('l')

# Generic variables for map values.
generic.variable('v')
generic.variable('w')


# ---------- Rewrite rules ---------- #
# isempty(empty) -> true.
kv_map.rewrite_rule(kv_map.isempty(kv_map.empty()),
                    # ->
                    boolean.true())

# isempty(empty, add(m, k, v)) -> false.
kv_map.rewrite_rule(kv_map.isempty(kv_map.add(kv_map.m(),
                                              generic.k(),
                                              generic.v())),
                    # ->
                    boolean.false())

# add(add(m, k, v), k, w) -> add(m, k, w).
kv_map.rewrite_rule(kv_map.add(kv_map.add(kv_map.m(),
                                          generic.k(),
                                          generic.v()),
                               generic.k(), generic.w()),
                    # ->
                    kv_map.add(kv_map.m(), generic.k(), generic.w()))

# equal(k, l) == false and isempty(m) == false
# => add(add(m, k, v), l, w) -> add(add(m, l, w), k, v).
kv_map.rewrite_rule(kv_map.add(kv_map.add(kv_map.m(),
                                          generic.k(),
                                          generic.v()),
                               generic.l(), generic.w()),
                    # ->
                    kv_map.add(kv_map.add(kv_map.m(),
                                          generic.l(),
                                          generic.w()),
                               generic.k(), generic.v()),
                    # Conditions:
                    [(generic.equal(generic.k(), generic.l()), boolean.false()),
                     (kv_map.isempty(kv_map.m()), boolean.false())])

# get(add(m, k, v), k) -> v.
kv_map.rewrite_rule(kv_map.get(kv_map.add(kv_map.m(),
                                          generic.k(),
                                          generic.v()),
                                generic.k()),
                    # ->
                    generic.v())

# equal(k, l) == false and isempty(m) == false
# => get(add(m, k, v), l) -> get(m, l).
kv_map.rewrite_rule(kv_map.get(kv_map.add(kv_map.m(),
                                          generic.k(),
                                          generic.v()),
                                generic.l()),
                    # ->
                    kv_map.get(kv_map.m(), generic.l()),
                    # Conditions:
                    [(generic.equal(generic.k(), generic.l()),
                      boolean.false()),
                     (kv_map.isempty(kv_map.m()), 
                      boolean.false())])

# equal(k, l) == false and isempty(m) == true
# => get(add(m, k, v), l) -> empty.
kv_map.rewrite_rule(kv_map.get(kv_map.add(kv_map.m(),
                                          generic.k(),
                                          generic.v()),
                                generic.l()),
                    # ->
                    kv_map.empty(),
                    # Conditions:
                    [(generic.equal(generic.k(), generic.l()),
                      boolean.false()),
                     (kv_map.isempty(kv_map.m()), 
                      boolean.true())])

# delete(add(m, k, v), k) -> m.
kv_map.rewrite_rule(kv_map.delete(kv_map.add(kv_map.m(),
                                             generic.k(),
                                             generic.v()),
                                  generic.k()),
                    # ->
                    kv_map.m())

# equal(k, l) == false and isempty(m) == false
# => delete(add(m, k, v), l) -> add(delete(m, l), k, v).
kv_map.rewrite_rule(kv_map.delete(kv_map.add(kv_map.m(),
                                             generic.k(),
                                             generic.v()),
                                  generic.l()),
                    # ->
                    kv_map.add(kv_map.delete(kv_map.m(), generic.l()),
                               generic.k(), generic.v()),
                    # Conditions:
                    [(generic.equal(generic.k(), generic.l()), 
                      boolean.false()),
                     (kv_map.isempty(kv_map.m()),
                      boolean.false())])

# equal(k, l) == false and isempty(m) == true
# => delete(add(m, k, v), l) -> add(delete(m, l), k, v).
kv_map.rewrite_rule(kv_map.delete(kv_map.add(kv_map.m(),
                                             generic.k(),
                                             generic.v()),
                                  generic.l()),
                    # ->
                    kv_map.m(),
                    # Conditions:
                    [(generic.equal(generic.k(), generic.l()),
                      boolean.false()),
                     (kv_map.isempty(kv_map.m()),
                      boolean.true())])



"""
Adt for the String Sort.
"""
# Aurelien Coet, 2018.

from alpynet.adt import Sort, GenericSort
from alpynet.adts.boolean import boolean


chars = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
          'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
          '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '_']


# ---------------------------------------- #
# Definition of the character sort.
# ---------------------------------------- #

char = Sort('char')


# ---------- Operations on characters ---------- #
# Generators.
for i, c in enumerate(chars):
    char.operation(c, ())
    if i < 26:
        char.operation(c.upper(), ())

# Observer.
char.operation('equal', (char, char), boolean)


# ---------- Character variables ---------- #
char.variable('c1')
char.variable('c2')


# ---------- Rewrite rules on characters ---------- #
for i, c in enumerate(chars):
    if i < 26:
        c_upper = c.upper()
    else:
        c_upper = None

    for j, c2 in enumerate(chars):
        if j < 26:
            c2_upper = c2.upper()
        else:
            c2_upper = None

        if c == c2:
            rhs = boolean.true()
        else:
            rhs = boolean.false()
        char.rewrite_rule(char.equal(char.__dict__[c](),
                                     char.__dict__[c2]()),
                          rhs)

        if c_upper:
            if c_upper == c2:
                rhs = boolean.true()
            else:
                rhs = boolean.false()
            char.rewrite_rule(char.equal(char.__dict__[c_upper](),
                                         char.__dict__[c2]()),
                              rhs)

        if c2_upper:
            if c == c2_upper:
                rhs = boolean.true()
            else:
                rhs = boolean.false()
            char.rewrite_rule(char.equal(char.__dict__[c](),
                                         char.__dict__[c2_upper]()),
                              rhs)

        if c_upper and c2_upper:
            if c_upper == c2_upper:
                rhs = boolean.true()
            else:
                rhs = boolean.false()
            char.rewrite_rule(char.equal(char.__dict__[c_upper](),
                                         char.__dict__[c2_upper]()),
                              rhs)


# ---------------------------------------- #
# Definition of the string sort.
# ---------------------------------------- #
string = Sort('str')


# ---------- Operations on strings ---------- #
# Generators.
string.operation('empty', ())
string.operation('append', (string, char))
string.operation('concat', (string, string))

# Observers.
string.operation('equal', (string, string), boolean)


# ---------- String variables ---------- #
string.variable('s')
string.variable('t')


# ---------- Rewrite rules on strings ---------- #
string.rewrite_rules += char.rewrite_rules

# equal(empty, empty) -> true.
string.rewrite_rule(string.equal(string.empty(), string.empty()),
                    boolean.true())

# equal(empty, append(s, c)) -> false.
string.rewrite_rule(string.equal(string.empty(),
                                 string.append(string.s(), char.c1())),
                    boolean.false())

# equal(append(s, c), empty) -> false.
string.rewrite_rule(string.equal(string.append(string.s(), char.c1()),
                                 string.empty()),
                    boolean.false())

# equal(append(s, c), append(t, c)) -> equal(s, t).
string.rewrite_rule(string.equal(string.append(string.s(), char.c1()),
                                 string.append(string.t(), char.c1())),
                    string.equal(string.s(), string.t()))

# equal(c, d) == false => equal(append(s, c), append(t, d)) -> false.
string.rewrite_rule(string.equal(string.append(string.s(), char.c1()),
                                 string.append(string.t(), char.c2())),
                    boolean.false(),
                    [(char.equal(char.c1(), char.c2()), boolean.false())])

generic = GenericSort()
generic.operation('equal', (generic, generic))

string.rewrite_rule(generic.equal(string.s(), string.t()),
                    string.equal(string.s(), string.t()))

# TODO: concat rewrite rules.



"""
Definition of exceptions.
"""

# APN exceptions.


class ConsumeException(Exception):
    """
    Exception raised when an attempt to consume a token in a place where it
    doesn't exist is made.
    """


class FiringException(Exception):
    """
    Exception raised when an attempt to fire a transition that is not fireable
    is made.
    """



"""
Adt for the Boolean Sort.
"""
# Aurelien Coet, 2018.

from alpynet.adt import Sort, GenericSort

# Sort Definition.
boolean = Sort('bool')

# ---------- Operations on booleans ---------- #
# Generators.
boolean.operation('true', ())
boolean.operation('false', ())

# Modifiers.
boolean.operation('not_', (boolean,))
boolean.operation('and_', (boolean, boolean))
boolean.operation('or_', (boolean, boolean))

# Observers.
boolean.operation('equal', (boolean, boolean))


# ---------- Variables ---------- #
boolean.variable('b')
boolean.variable('c')


# ---------- Rewrite rules ---------- #
# not(true) -> false.
boolean.rewrite_rule(boolean.not_(boolean.true()), boolean.false())

# not(false) -> true.
boolean.rewrite_rule(boolean.not_(boolean.false()), boolean.true())

# and(b, false) -> false.
boolean.rewrite_rule(boolean.and_(boolean.b(), boolean.false()),
                     boolean.false())

# and(b, true) -> b.
boolean.rewrite_rule(boolean.and_(boolean.b(), boolean.true()),
                     boolean.b())

# or(b, false) -> b.
boolean.rewrite_rule(boolean.or_(boolean.b(), boolean.false()),
                     boolean.b())

# or(b, true) -> true.
boolean.rewrite_rule(boolean.or_(boolean.b(), boolean.true()),
                     boolean.true())

generic = GenericSort()
generic.operation('equal', (generic, generic))

boolean.rewrite_rule(generic.equal(boolean.b(), boolean.c()),
                     boolean.equal(boolean.b(), boolean.c()))



"""
Model for Algebraic Petri Nets (APNs).
"""
# Aurelien Coet, 2018.

import random
import graphviz as gv
from alpynet.adt import Sort, Term, RewriteRule
from alpynet.exceptions import ConsumeException, FiringException


class Place(object):
    """
    Place in an algebraic petri net (APN).
    """

    def __init__(self, name, sort, marking=[]):
        assert type(name) == str, "Name of a place must be a string"
        assert isinstance(sort, Sort), "Sort of a place must be a sort"
        assert type(marking) == list, "Initial marking must be a list"
        for token in marking:
            assert isinstance(token, Term), "Tokens in a place must be terms"
            assert token.sort == sort, "Tokens' sorts must match the place's"
        self.name = name
        self.sort = sort
        self.marking = marking

    def __str__(self):
        return "place {}".format(self.name)

    def consume(self, tokens):
        """
        Consume tokens from the place.

        Args:
            tokens: A list of tokens to consume from the place.
        """
        assert type(tokens) == list, "Tokens must be a list of terms"
        new_marking = self.marking.copy()
        for token in tokens:
            assert isinstance(token, Term), "Tokens must be terms"
            if token not in self.marking:
                raise ConsumeException
            new_marking.remove(token)
        self.marking = new_marking

    def produce(self, tokens):
        """
        Produce tokens in the place.

        Args:
            tokens: A list of tokens to produce in the place.
        """
        assert type(tokens) == list, "Tokens must be a list of terms"
        new_marking = self.marking.copy()
        for token in tokens:
            assert isinstance(token, Term), "Tokens must be terms"
            assert token.sort == self.sort, "Tokens must have the place's sort"
            new_marking.append(token)
        self.marking = new_marking


class Transition(object):
    """
    Transition in an algebraic petri net (APN).
    """

    def __init__(self, name):
        assert type(name) == str, "Name of a transition must be a string"
        self.name = name
        self.inbound_arcs = []
        self.outbound_arcs = []

    def __str__(self):
        return "transition {}".format(self.name)

    def inbound_arc(self, source, label):
        """
        Add an arc to the list of inbound arcs of the transition.

        Args:
            arc: The arc to add to the list.
        """
        assert isinstance(source, Place), "Source of arc must be a place"
        assert type(label) == list, "Label must be a list of terms"
        for term in label:
            assert isinstance(term, Term), "Elements in label must be terms"
        self.inbound_arcs.append(Arc(source, self, label))

    def outbound_arc(self, target, label):
        """
        Add an arc to the list of outbound arcs of the transition.

        Args:
            arc: The arc to add to the list.
        """
        assert isinstance(target, Place), "Target of arc must be a place"
        assert type(label) == list, "Label must be a list of terms"
        for term in label:
            assert isinstance(term, Term), "Elements in label must be terms"
        self.outbound_arcs.append(Arc(self, target, label))

    def fireable(self):
        """
        Check if the transition is fireable, and if so, compute the variable
        bindings that make it possible.

        Returns:
            A boolean indicating whether the transition can be fired,
            as well as the bindings for the variables in the labels
            of its inbound and outbound arcs if it can be fired.
        """
        bindings = {}
        matched_tokens = []

        for arc in self.inbound_arcs:
            has_match = False

            for term in arc.label:
                for token in arc.source.marking:
                    # If a token in a precondition is already being consumed by
                    # some other term on an inbound arc, it cannot be consumed
                    # a second time.
                    if token in matched_tokens:
                        continue

                    (matching, binding) = term.match(token)
                    if matching:
                        has_match = True
                        for key, value in binding.items():
                            # If there is some conflict in the variable
                            # bindings, the transition cannot be fired.
                            if key in bindings and bindings[key] != value:
                                has_match = False
                            else:
                                bindings[key] = value
                                matched_tokens.append(term)
                        break

                if has_match is False:
                    return (False, {})

        return (True, bindings)

    def _consume_inbound(self, bindings, rewrite_rules):
        """
        Consume tokens from the places connected to the inbound arcs of the
        transition.

        Args:
            bindings: A dict of variable bindings to be applied to the terms
                on the labels of the inbound arcs of the transition.
            rewrite_rules: A list of rewrite rules to be used to reduce the
                terms on the labels of the inbound arcs of the transition.
        """
        for arc in self.inbound_arcs:
            tokens = []
            for term in arc.label:
                tokens.append(term.apply_binding(bindings)
                                  .reduce(rewrite_rules))
            arc.source.consume(tokens)

    def _produce_outbound(self, bindings, rewrite_rules):
        """
        Produce tokens in the places connected to the outbound arcs of the
        transition.

        Args:
            bindings: A dict of variable bindings to be applied to the terms
                on the labels of the outbound arcs of the transition.
            rewrite_rules: A list of rewrite rules to be used to reduce the
                terms on the labels of the outbound arcs of the transition.
        """
        for arc in self.outbound_arcs:
            tokens = []
            for term in arc.label:
                tokens.append(term.apply_binding(bindings)
                                  .reduce(rewrite_rules))
            arc.target.produce(tokens)

    def fire(self, rewrite_rules=[]):
        """
        Fire the transition. Consumes tokens in the preconditions of the
        transition, and produces new ones in its postconditions. If there
        are statements inside the transition, those are evaluated, and
        their associated contexts are updated accordingly in the variable
        bindings for the arcs of the transition.

        Args:
            rewrite_rules: A list of rewrite rules to use to reduce the
                terms on the inbound and outbound arcs of the transition.

        Raises:
            A FiringException when the transition cannot be fired.
        """
        fireable, bindings = self.fireable()

        if fireable is False:
            raise FiringException

        self._consume_inbound(bindings, rewrite_rules)
        self._produce_outbound(bindings, rewrite_rules)


class Arc(object):
    """
    Arc between places and transitions in an algebraic petri net (APN).
    """

    def __init__(self, source, target, label):
        assert isinstance(source, Place) or isinstance(source, Transition),\
            "Source of an arc must be a place or transition"
        assert isinstance(target, Place) or isinstance(target, Transition),\
            "Target of an arc must be a place or transition"
        assert type(label) == list, "Label of an arc must be a list of terms"
        for term in label:
            assert isinstance(term, Term), "Content of the label must be terms"
        self.source = source
        self.target = target
        self.label = label

    def __str__(self):
        txt = "arc from {} to {}, with label [".format(self.source,
                                                       self.target)
        for term in self.label:
            txt += str(term) + ', '
        txt += "]"
        return txt


class AlgebraicPetriNet(object):
    """
    Algebraic Petri Net (APN).
    """

    def __init__(self, name, places=[], transitions=[], rewrite_rules=[]):
        assert type(name) == str, "Name of an APN must be a string"
        assert type(places) == list, "Places must be a list of Places"
        for place in places:
            assert isinstance(place, Place),\
                "Places must be instances of Place"
        assert type(transitions) == list,\
            "Transitions must be a list of transitions"
        for transition in transitions:
            assert isinstance(transition, Transition),\
                "Transitions must be instances of Transition"
        for rule in rewrite_rules:
            assert isinstance(rule, RewriteRule),\
                "Rewrite rules in the APN must be instances of RewriteRule"
        self.name = name
        self.places = places
        self.transitions = transitions
        self.rewrite_rules = rewrite_rules

    def __str__(self):
        return "Algebraic Petri Net {}".format(self.name)

    def add_place(self, name, sort, marking=[]):
        """
        Add a place to the APN.

        Args:
            name: The name of the place.
            sort: The sort of the place.
            marking: A list of terms representing the initial marking of the
                place.

        Returns:
            The newly created place.
        """
        place = Place(name, sort, marking)
        self.places.append(place)
        return place

    def add_transition(self, name):
        """
        Add a transition to the APN.

        Args:
            name: The name of the transition.

        Returns:
            The newly created transition.
        """
        transition = Transition(name)
        self.transitions.append(transition)
        return transition

    def add_arc(self, source, target, label=[]):
        """
        Add an arc to the APN.

        Args:
            source: The source of the arc (a Place or a Transition).
            target: The target of the arc (a Place or a Transition).
            label: The label for the arc (a list of terms).
        """
        assert isinstance(source, Place) or isinstance(source, Transition),\
            "Source of an arc must be a Place or Transition"
        assert isinstance(target, Place) or isinstance(target, Transition),\
            "Target of an arc must be a Place or transition"
        assert type(source) != type(target),\
            "Source and target of an arc cannot be both Places or Transitions"
        if isinstance(source, Place):
            assert source in self.places, "Source must exist in the APN"
            assert target in self.transitions, "Target must exist in the APN"
            target.inbound_arc(source, label)
        else:
            assert source in self.transitions, "Source must exist in the APN"
            assert target in self.places, "Target must exist in the APN"
            source.outbound_arc(target, label)

    def marking(self):
        """
        Get the markings of the places of the APN.

        Returns:
            A dict with the places of the APN as keys and their markings as
            values.
        """
        markings = {}
        for place in self.places:
            markings[place] = place.marking
        return markings

    def fireables(self):
        """
        Get the list of transitions that are fireable in the APN given its
        current marking.

        Returns:
            A list of transitions that can be fired.
        """
        fireables = []
        for transition in self.transitions:
            fireable, _ = transition.fireable()
            if fireable:
                fireables.append(transition)
        return fireables

    def fire(self, transition):
        """
        Fire a transition in the APN.

        Args:
            transition: The transition to fire in the APN.
        """
        assert transition in self.transitions, "Transition must be in the APN"
        transition.fire(self.rewrite_rules)

    def fire_random(self):
        """
        Randomly fire one of the fireable transitions of the APN.
        """
        random.choice(self.fireables()).fire(self.rewrite_rules)

    def visualise(self, filepath, format='pdf'):
        """
        Visualise the APN and its current marking graphically.

        Args:
            filepath: The path to the file where the visualisation of the APN
                must be saved.
            format: The format in which the visualisation must be saved.
        """
        graph = gv.Digraph(name=self.name, format=format,
                           graph_attr={'label': self.name})

        for place in self.places:
            label = ""
            for token in place.marking:
                label += str(token) + '\n'
            graph.node(place.name, label=label,
                       _attributes={'forcelabels': 'true',
                                    'xlabel': place.name})

        for transition in self.transitions:
            graph.node(transition.name, label=transition.name,
                       _attributes={'shape': 'box'})

            for arc in transition.inbound_arcs:
                label = ""
                for term in arc.label:
                    label += str(term) + '\n'
                graph.edge(arc.source.name, arc.target.name, label=label)

            for arc in transition.outbound_arcs:
                label = ""
                for term in arc.label:
                    label += str(term) + '\n'
                graph.edge(arc.source.name, arc.target.name, label=label)

        graph.render(filepath, view=True)



"""
Model for Algebraic Data Types (ADTs) and terms.
"""
# Aurelien Coet, 2018.


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls)\
                                    .__call__(*args, **kwargs)
        return cls._instances[cls]


class Sort(object):
    """
    Sort in an ADT.

    A sort has operations and variables defined on it. Those can be used to
    define terms of the sort. It also has rewrite rules associated to it.
    """

    def __init__(self, name):
        assert type(name) == str, "Name of a sort must be a string"
        self.name = name
        self.rewrite_rules = []

    def __str__(self):
        return self.name

    def operation(self, name, signature, sort=None):
        """
        Create a new operation on the sort. The operation is stored in
        the sort's attributes.

        Args:
            name: The name of the operation.
            signature: The signature of the operation (the list of sorts it
                       takes as input).
            sort: The sort of the result of the operation. Defaults to self
                (the sort to which the operation is attached).
        """
        assert type(name) == str, "Name of an operation must be a string"
        assert type(signature) == tuple, "Signature must be a tuple of sorts"
        for s in signature:
            assert isinstance(s, Sort), "Signature must be a tuple of sorts"
        if sort is None:
            sort = self
        assert isinstance(sort, Sort), "Sort of the operation must be a sort"
        self.__dict__[name] = Operation(name, signature, sort)

    def variable(self, name):
        """
        Create a new variable of the sort represented by the object. The
        variable is stored in the sort's attributes.

        Args:
            name: The name of the variable.
        """
        assert type(name) == str, "Name of a variable must be a string"
        self.__dict__[name] = Variable(name, self)

    def rewrite_rule(self, lhs, rhs, conditions=[]):
        """
        Create a new rewrite rule for the sort.

        Args:
            lhs: The left hand side of the rule.
            rhs: The right hand side of the rule.
            condition: The condition of the rule.
        """
        assert isinstance(lhs, Term),\
            "Left hand side of the rule must be a term"
        assert isinstance(rhs, Term),\
            "Right hand side of the rule must be a term"
        assert type(conditions) == list,\
            "Conditions must be a list of conditions"
        for condition in conditions:
            assert len(condition) == 2, "Conditions must contain 2 terms"
            assert isinstance(condition[0], Term),\
                "Conditions must contain terms"
            assert isinstance(condition[1], Term),\
                "Conditions must contain terms"
        self.rewrite_rules.append(RewriteRule(lhs, rhs, conditions))


class GenericSort(Sort, metaclass=Singleton):
    """
    Generic sort in an ADT.

    Represents any sort. Shouldn't be used to define a specific sort with
    operations and variables. It is just a placeholder for any other sort in
    the definition of an operation's arguments and to define terms that can
    be of any sort.
    """

    def __init__(self):
        super().__init__('anysort')


class Operation(object):
    """
    Operation in an ADT.

    An operation is similar to a function that takes arguments of
    given sorts as input and returns a value of a given sort.
    """

    def __init__(self, name, signature, sort):
        assert type(name) == str, "Name of an operation must be a string"
        assert type(signature) == tuple,\
            "Signature of an operation must be a tuple of sorts"
        for s in signature:
            assert isinstance(s, Sort),\
                "Signature of an operation must be a tuple of sorts"
        assert isinstance(sort, Sort),\
            "Sort of the operation must be an instance of Sort"
        self.name = name
        self.signature = signature
        self.sort = sort

    def __eq__(self, other):
        if type(other) != Operation:
            return False

        for attr, value in self.__dict__.items():
            if value != other.__dict__[attr]:
                return False

        return True

    def __str__(self):
        txt = "{}.{}(".format(self.sort, self.name)
        for s in self.signature:
            txt += "{}, ".format(str(s))
        txt += ") -> {}".format(self.sort)
        return txt

    def __repr__(self):
        return str(self)

    def __call__(self, *args):
        return Term(self, args)


class Variable(object):
    """
    Variable of a given sort in an ADT.

    A variable is an expression that can be replaced by any value of its sort
    in a term.
    """

    def __init__(self, name, sort):
        assert type(name) == str, "Name of a variable must be a string"
        assert isinstance(sort,  Sort),\
            "Sort associated to a variable must be an instance of Sort"
        self.name = name
        self.sort = sort

    def __str__(self):
        return "{}.{}".format(str(self.sort), self.name)

    def __repr__(self):
        return str(self)

    def __call__(self):
        return Term(self)


class Term(object):
    """
    Term in an ADT.
    """

    def __init__(self, head, args=()):
        assert isinstance(head, Operation) or isinstance(head, Variable),\
            "Head of a term must be a variable or an operation"
        if type(head) == Operation:
            assert type(args) == tuple,\
                "Arguments of an operation in a term must be a tuple of terms"
            for i, arg in enumerate(args):
                assert isinstance(arg, Term),\
                    "Arguments of an operation must be terms"
                if type(head.signature[i]) != GenericSort:
                    assert arg.sort == head.signature[i],\
                        "Arguments sorts must match the head's signature"
        else:
            assert len(args) == 0, "A variable cannot have arguments"

        self.head = head
        self.sort = head.sort
        self.args = args

    def __eq__(self, other):
        if type(other) != Term or other.head != self.head:
            return False

        for i, operand in enumerate(self.args):
            if operand != other.args[i]:
                return False
        return True

    def __neq__(self, other):
        return not self == other

    def __str__(self):
        if type(self.head) == Operation:
            txt = "{}.{}(".format(str(self.head.sort), self.head.name)
            for i, arg in enumerate(self.args):
                txt += "{}".format(str(arg))
                if i < len(self.args)-1:
                    txt += ", "
            txt += ")"
            return txt
        return str(self.head)

    def __repr__(self):
        return str(self)

    def match(self, other):
        """
        Check if two terms match and compute the corresponding variable
        bindings if they do.

        Args:
            other: The term to match this one with.

        Returns:
            A tuple containing a boolean indicating whether the terms
            match, as well as a dict with the variables bindings if
            they do (the dict is empty if they don't).
        """
        assert isinstance(other, Term), "Other must be another term"

        if type(self.sort) != GenericSort and\
           type(other.sort) != GenericSort and\
           self.sort != other.sort:
            return (False, {})

        if type(self.sort) == GenericSort and\
           type(self.head) == Operation and\
           self.head.name not in other.head.sort.__dict__:
            return (False, {})

        if type(other.sort) == GenericSort and\
           type(other.head) == Operation and\
           other.head.name not in self.head.sort.__dict__:
            return (False, {})

        bindings = {}

        def compare(lhs, rhs):
            if type(lhs.head) == Variable and type(rhs.head) == Variable:
                if (lhs.head in bindings and bindings[lhs.head] != rhs)\
                   or (rhs.head in bindings and bindings[rhs.head] != lhs):
                    return False
                else:
                    bindings[lhs.head] = rhs
                    bindings[rhs.head] = lhs
                    return True

            elif type(lhs.head) == Variable and type(rhs.head) == Operation:
                if lhs.head in bindings and bindings[lhs.head] != rhs:
                    return False
                else:
                    bindings[lhs.head] = rhs
                    return True

            elif type(lhs.head) == Operation and type(rhs.head) == Variable:
                if rhs.head in bindings and bindings[rhs.head] != lhs:
                    return False
                else:
                    bindings[rhs.head] = lhs
                    return True

            elif type(lhs.head) == Operation and type(rhs.head) == Operation:
                if lhs.head == rhs.head:
                    equivalent = True
                    for i in range(len(lhs.args)):
                        equivalent = equivalent and compare(lhs.args[i],
                                                            rhs.args[i])
                    return equivalent
                else:
                    return False

            else:
                return False

        return (compare(self, other), bindings)

    def apply_binding(self, binding):
        """
        Apply a set of variable bindings on the variables of a term.

        Args:
            binding: The set of variable bindings to apply on the term.

        Returns:
            A new term where the variables are replaced by their bindings.
        """
        assert type(binding) == dict, "Binding must be a dict"
        for key, value in binding.items():
            assert isinstance(key, Variable),\
                "Keys in binding must be variables"
            assert isinstance(value, Term),\
                "Variable bindings must be terms"

        def rename(t):
            if type(t.head) == Variable:
                return binding[t.head] or t
            else:
                args = []
                for arg in t.args:
                    args.append(rename(arg))
                return Term(t.head, tuple(args))

        return rename(self)

    def reduce(self, rewrite_rules):
        """
        Reduce the term by applying a set of rewrite rules on it until a
        fixpoint is reached. The resulting term is in normal form
        (accordingly to the rewrite rules passed as argument).

        Args:
            rewrite_rules: A list of rewrite rules to be applied on the term
                in order to reduce it to its normal form.

        Returns:
            A new term obtained after applying the rewrite rules on the term
            until a fixpoint (normal form) was reached.
        """
        prev_term = None
        new_term = Term(self.head, self.args)
        while new_term != prev_term:
            prev_term = new_term
            for rule in rewrite_rules:
                new_term = rule.apply(new_term, rewrite_rules)

        return new_term


class RewriteRule(object):
    """
    Rewrite rule for terms in ADTs.
    """

    def __init__(self, lhs, rhs, conditions=[]):
        assert isinstance(lhs, Term), "Left hand side must be a term"
        assert isinstance(rhs, Term), "Right hand side must be a term"
        assert type(conditions) == list,\
            "Conditions must be a list of conditions"
        for condition in conditions:
            assert len(condition) == 2, "Conditions must contain 2 terms"
            assert isinstance(condition[0], Term),\
                "Conditions must contain terms"
            assert isinstance(condition[1], Term),\
                "Conditions must contain terms"
        self.lhs = lhs
        self.rhs = rhs
        self.conditions = conditions

    def __str__(self):
        txt = ""
        if self.conditions:
            for condition in self.conditions: 
                txt += "({} == {}), ".format(str(condition[0]),
                                             str(condition[1]))
            txt += "=> "
        txt += "{} -> {}".format(str(self.lhs), str(self.rhs))
        return txt

    def __repr__(self):
        return str(self)

    def apply(self, term, rewrite_rules=[]):
        """
        Apply the rewrite rule on a term if it is possible.

        Args:
            term: The term to apply the rule on.
            rewrite_rules: A list of rewrite rules to use to reduce the
                conditions of the rule.

        Returns:
            A new term with where the rewrite rule has been applied.
        """
        # Recursive application of the rule on the arguments of the term
        # -> left-right innermost strategy.
        args = []
        for arg in term.args:
            args.append(self.apply(arg))

        new_term = Term(term.head, tuple(args))

        matching, binding = new_term.match(self.lhs)
        if matching:
            if self.conditions:
                for condition in self.conditions:
                    if condition[0].apply_binding(binding)\
                                   .reduce(rewrite_rules) !=\
                       condition[1].apply_binding(binding)\
                                   .reduce(rewrite_rules):
                        return new_term
            return Term(self.rhs.head, self.rhs.args).apply_binding(binding)
        else:
            return new_term



"""
Adt for the Natural Sort.
"""
# Aurelien Coet, 2018.

from alpynet.adt import Sort, GenericSort
from alpynet.adts.boolean import boolean


# Sort Definition.
nat = Sort('nat')


# ---------- Operations on naturals ---------- #
# Generators.
nat.operation('zero', ())
nat.operation('succ', (nat,))

# Modifiers.
nat.operation('add', (nat, nat))

# Observers.
nat.operation('equal', (nat, nat), boolean)


# ---------- Variables ---------- #
nat.variable('x')
nat.variable('y')


# ---------- Rewrite rules ---------- #
# add(x, 0) -> x.
nat.rewrite_rule(nat.add(nat.x(), nat.zero()), nat.x())

# add(x, succ(y)) -> succ(add(x, y)).
nat.rewrite_rule(nat.add(nat.x(), nat.succ(nat.y())),
                 nat.succ(nat.add(nat.x(), nat.y())))

# equal(0, 0) -> true.
nat.rewrite_rule(nat.equal(nat.zero(), nat.zero()),
                 boolean.true())

# equal(succ(x), 0) -> false.
nat.rewrite_rule(nat.equal(nat.succ(nat.x()), nat.zero()),
                 boolean.false())

# equal(0, succ(x)) -> false.
nat.rewrite_rule(nat.equal(nat.zero(), nat.succ(nat.x())),
                 boolean.false())

# equal(succ(x), succ(y)) -> equal(x, y).
nat.rewrite_rule(nat.equal(nat.succ(nat.x()), nat.succ(nat.y())),
                 nat.equal(nat.x(), nat.y()))

generic = GenericSort()
generic.operation('equal', (generic, generic))

nat.rewrite_rule(generic.equal(nat.x(), nat.y()),
                 nat.equal(nat.x(), nat.y()))



from setuptools import setup


setup(name='alpynet',
      version=1.0,
      url='https://github.com/coetaur0/alpynet',
      license='Apache 2',
      author='Aurelien Coet',
      author_email='aurelien.coet19@gmail.com',
      description='Algebraic Petri Nets in Python3',
      packages=[
        'alpynet',
        'alpynet.adts',
      ],
      test_suite='tests',
      install_requires=[
        'graphviz',
      ]
      )
